"""
A non-linear curve fitting example for partial discharges.
"""

from artap.problem import Problem
from artap.results import Results
# from artap.algorithm_bayesopt import BayesOptSerial
from artap.algorithm_nlopt import NLopt, LN_BOBYQA, LN_COBYLA, LN_NELDERMEAD
from artap.algorithm_genetic import NSGAII
from artap.algorithm_swarm import PSO
import matplotlib.pyplot as plt
import numpy as np

# Data Samples from the test examples (page 20 in the bsc theses)
# n = 1
C1x = [9.627, 19.589]
R1x = [5.870, 2.1663]

# (tdchi, SRi)
SR1 = [(1.0, 46.83347), (2.0, 45.74274), (4.0, 43.6369), (6.0, 41.62801), (8.0, 39.7116), (10.0, 37.88341),
       (15.0, 33.67273), (20.0, 29.93007), (30.0, 23.64647), (50.0, 14.75991), (75.0, 8.188986), (100.0, 4.543356),
       (150.0, 1.398524), (200.0, 0.43049), (300.0, 0.04079), (500.0, 0.000366)]

# n = 2
C2x = [9.627, 10.236, 9.4081]
R2x = [5.870, 2.1663, 1.7806]

C2o1 = [9.627, 9.958385741447383, 9.68637611277249]
R2o1 = [5.870, 1.692170503816123,  2.3120386928356145] # from the newton method based search

C2o2 = [9.627, 2.355104843390313, 17.3250807222832]
R2o2 = [5.870, 13.017148342665315, 1.0569093531258873]


# (tdchi, SRi)
SR2 = [(1.0, 100.7923), (2.0, 95.58673), (4.0, 85.98188), (6.0, 77.35857), (8.0, 69.6149), (10.0, 62.65971),
       (15.0, 48.20685), (20.0, 37.13707), (30.0, 22.12758), (50.0, 7.979435), (75.0, 2.29236), (100.0, 0.676797),
       (150.0, 0.062897), (200.0, 0.062897), (300.0, 6.49E-05), (500.0, 7.75E-09)]

# n = 3
C3x = [9.627, 10.236, 9.4081, 46.288]
R3x = [5.870, 2.1663, 1.7806, 3.0078]

# (tdchi, SRi)
SR3 = [(1.0, 135.0715), (2.0, 129.6215), (4.0, 119.5329), (6.0, 110.4325), (8.0, 102.2183), (10.0, 94.79905),
       (15.0, 79.21414), (20.0, 67.05153), (30.0, 49.96951), (50.0, 32.09568), (75.0, 22.44427), (100.0, 17.51621),
       (150.0, 11.8215), (200.0, 8.217025), (300.0, 4.003668), (500.0, 0.951871)]

Uch = 1000.  # V
tch = 4000.  # s
c0 = 9.627


def sri(x, tdch, Uch, tch, c0):
    """
    Calculates the ui voltage from the given parameters
    :param x: list of individuals from the optimized paramters in the ri, ci order
    :param Sri: measured data at the ith case, contains tuples : (tdchi, Sri) format
    :return:
    """
    sri = 0.
    for j in range(2, len(x), 2):  # step is 2, because r,c pairs
        taui = x[j] * x[j + 1]  # ri*ci
        sri += Uch * (1.0 - np.exp(-tch / taui)) * np.exp(-tdch / taui) / x[j] / c0
    return sri


def test_ui():
    # C1x = [9.627, 19.589]
    # R1x = [5.870, 2.1663]

    # shape of the x  in the case of n = 1:
    # x = [r0,c0,r1,c1]

    # in the simplest testproblem
    # x = [R1x[0], C1x[0], R1x[1], C1x[1]]
    #
    # for sr in SR1:
    #     print('measured:',sr[1], 'calculated',sri(x, sr[0], Uch, tch, c0), 'error:', sr[1]-sri(x, sr[0], Uch, tch, c0))

    # tests for n = 2
    x = [R2x[0], C2x[0], R2x[1], C2x[1], R2x[2], C2x[2]]

    err = 0.0
    for sr in SR2:
        err += (sr[1] - sri(x, sr[0], Uch, tch, c0))**2.
        #print('measured:',sr[1], 'calculated',sri(x, sr[0], Uch, tch, c0), 'error:', sr[1]-sri(x, sr[0], Uch, tch, c0))
    print("standard error:", err)

    # x2 = [R2o1[0], C2o1[0], R2o1[1], C2o1[1], R2o1[2], C2o1[2]]
    # err = 0.0
    # for sr in SR2:
    #     err += (sr[1] - sri(x2, sr[0], Uch, tch, c0)) ** 2.
    #     # print('measured:',sr[1], 'calculated',sri(x, sr[0], Uch, tch, c0), 'error:', sr[1]-sri(x, sr[0], Uch, tch, c0))
    # print("standard error:", err)
    # tests for n = 3
    # x = [R3x[0], C3x[0], R3x[1], C3x[1], R3x[2], C3x[2], R3x[3], C3x[3]]
    #
    # for sr in SR3:
    #     print('measured:', sr[1], 'calculated', sri(x, sr[0], Uch, tch, c0), 'error:',
    #           sr[1] - sri(x, sr[0], Uch, tch, c0))


class ExponentialFittingSimple(Problem):

    def set(self):
        self.name = 'Exponential fitting Problem'
        self.parameters = [{'name': 'r_1', 'bounds': [1.0, 20.]},
                           {'name': 'c_1', 'bounds': [1.0, 20.]},
                           {'name': 'r_2', 'bounds': [1.0, 20.]},
                           {'name': 'c_2', 'bounds': [1.0, 20.]}]

        self.costs = [{'name': 'f_1', 'criteria': 'minimize'}]

    def evaluate(self, individual):
        # The individual.vector function contains the problem parameters in the appropriate (previously defined) order

        x = [5.870, 9.627] # r0, C0 because its known
        x += individual.vector
        #print(x)
        error = 0.0

        # we just minimizing the square of the error
        for sr in SR2:
            error += (sr[1] - sri(x, sr[0], Uch, tch, c0))**2.

        m = len(SR2)
        #aic = m*(np.log(error/m))+4.
        return [np.log(error)]


#test_ui()
# Initialization of the problem

def nsgaii():
    problem = ExponentialFittingSimple()

    # Perform the optimization iterating over 100 times on 100 individuals.
    algorithm = NSGAII(problem)
    algorithm.options['max_population_number'] = 200
    algorithm.options['max_population_size'] = 100
    algorithm.run()

    results = Results(problem)
    optimum = results.find_minimum('f_1')
    print(optimum)
    x = [5.870, 9.627]  # r0, C0 because its known
    x += optimum.vector
    print(x)
    err = 0.0
    for sr in SR2:
        err += (sr[1] - sri(x, sr[0], Uch, tch, c0))**2.
    print("standard error, now:", err)


nsgaii()
test_ui()





from artap.problem import Problem
from artap.benchmark_functions import Booth
from artap.surrogate import SurrogateModelEval, SurrogateModelScikit


from sklearn.svm import SVR
from sklearn.neural_network import MLPRegressor
from sklearn.linear_model import SGDRegressor
from sklearn.tree import DecisionTreeRegressor, ExtraTreeRegressor
from sklearn.ensemble import AdaBoostRegressor, GradientBoostingRegressor, RandomForestRegressor, ExtraTreesRegressor, BaggingRegressor
from sklearn.kernel_ridge import KernelRidge
from sklearn.neighbors import KNeighborsRegressor, RadiusNeighborsRegressor
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF, Matern, DotProduct, WhiteKernel, ConstantKernel, RationalQuadratic, ExpSineSquared


class ProblemBooth(Problem):
    def __init__(self, name):
        parameters = {'x_1': {'initial_value': 2.5, 'bounds': [-10, 10]},
                      'x_2': {'initial_value': 1.5, 'bounds': [-10, 10]}}

        # costs = {'F': {'type': Problem.MINIMIZE, 'value': 0.0}}
        costs = ['F']

        super().__init__(name, parameters, costs)

        # enable surrogate
        self.surrogate = SurrogateModelScikit(self)

    def evaluate(self, x):
        return [Booth.eval(x)]


def output(regressor):
    xmeas = [[2.5, 1.5], [7.5, 1.5], [2.5, 6.5], [-2.5, 1.5], [2.5, -3.5], [2.2, 1.8], [2.69898454, 1.83185008], [2.23185008, 1.30101546], [1.98284599, 1.98211189], [1.94231886, 2.0199368]]
    ymeas = [[4.5], [144.5], [114.5], [114.5], [144.5], [2.88], [5.37826428], [5.27686152], [2.00699423], [1.85418726]]

    problem = ProblemBooth("ProblemBooth")
    problem.surrogate = SurrogateModelScikit(problem)
    problem.surrogate.sigma_threshold = 0.1
    problem.surrogate.train_step = len(xmeas)
    problem.surrogate.x_data = xmeas
    problem.surrogate.y_data = ymeas

    p = [1.04, 3.02]
    val = problem.evaluate(p)

    problem.surrogate.regressor = regressor
    problem.surrogate.train()
    res = problem.surrogate.predict(p)

    # print(problem.surrogate.regressor)
    print("optimum: value = {}, predict = {}, class = {}".format(val, res, problem.surrogate.regressor.__class__.__name__))


# Ensemble Methods
# AdaBoost regressor
output(AdaBoostRegressor(DecisionTreeRegressor(max_depth=4), n_estimators=300))
# Bagging regressor
output(BaggingRegressor(n_estimators=10))
# Gradient Boosting for regression
output(GradientBoostingRegressor(n_estimators=500, max_depth=4, min_samples_split=2, learning_rate=0.01, loss='ls'))
# Ensemble Methods - decision trees
output(ExtraTreesRegressor(n_estimators=10))
output(RandomForestRegressor(n_estimators=10))
# Decision Trees
# Extremely randomized tree regressor
output(ExtraTreeRegressor())
# Decision tree regressor
output(DecisionTreeRegressor())

# Gaussian Processes
output(GaussianProcessRegressor(kernel=1.0 * RBF(length_scale=1.0, length_scale_bounds=(1e-1, 10.0)), n_restarts_optimizer=9))
output(GaussianProcessRegressor(kernel=1.0 * ExpSineSquared(length_scale=1.0, periodicity=3.0, length_scale_bounds=(0.1, 10.0), periodicity_bounds=(1.0, 10.0)), n_restarts_optimizer=9))
output(GaussianProcessRegressor(kernel=1.0 * Matern(length_scale=1.0, length_scale_bounds=(1e-5, 1e5), nu=1.5), n_restarts_optimizer=9))
output(GaussianProcessRegressor(kernel=ConstantKernel(1.0, (1e-3, 1e3)) * RBF(10, (1e-2, 1e2)), n_restarts_optimizer=9))

# Generalized Linear Models
output(SGDRegressor())

# Nearest Neighbors
# Regression based on k-nearest neighbors
output(KNeighborsRegressor(n_neighbors=6, weights='distance'))
# Regression based on neighbors within a fixed radius
output(RadiusNeighborsRegressor())

output(KernelRidge(alpha=1.0))
output(SVR(kernel='rbf', C=1e4, gamma=0.1))

# Neural network models
# Multi-layer Perceptron regressor
output(MLPRegressor(hidden_layer_sizes=(6), activation='logistic', solver='lbfgs'))
output(MLPRegressor(hidden_layer_sizes=(10,),  activation='relu', solver='adam', alpha=0.001, batch_size='auto',
                    learning_rate='constant', learning_rate_init=0.01, power_t=0.5, max_iter=1000, shuffle=True,
                    random_state=9, tol=0.0001, verbose=False, warm_start=False, momentum=0.9, nesterovs_momentum=True,
                    early_stopping=False, validation_fraction=0.1, beta_1=0.9, beta_2=0.999, epsilon=1e-08))



from .problem import Problem
from .algorithm import Algorithm
from .job import Job

import warnings
import numpy as np
from scipy.stats import norm
from scipy.optimize import minimize

from sklearn.gaussian_process.kernels import Matern
from sklearn.gaussian_process import GaussianProcessRegressor

"""
Parts from 

author='Fernando Nogueira'
https://github.com/fmfn/BayesianOptimization'
author_email="fmfnogueira@gmail.com"
description='A Python implementation of global optimization with gaussian processes.'
"""


class Bayes(Algorithm):
    """ BayesOpt algorithms """

    def __init__(self, problem: Problem, name="BayesOpt"):
        super().__init__(problem, name)
        self.problem = problem

        self.options.declare(name='l_type', default='L_EMPIRICAL', values=_l_type,
                             desc='Type of learning for the kernel params')
        self.options.declare(name='sc_type', default='SC_MAP', values=_sc_type,
                             desc='Type of learning for the kernel params')
        self.options.declare(name='n_iterations', default=50, lower=1,
                             desc='Maximum BayesOpt evaluations')
        self.options.declare(name='init_method', default=1,
                             desc='Init method') # 1-LHS, 2-Sobol
        self.options.declare(name='n_init_samples', default=10, lower=1,
                             desc='Number of samples before optimization')
        self.options.declare(name='n_iter_relearn', default=10, lower=1,
                             desc='Number of samples before relearn kernel')
        self.options.declare(name='surr_name', default='sGaussianProcessML', values=_surr_name,
                             desc='Name of the surrogate function')
        self.options.declare(name='surr_noise', default=1e-10, lower=0.0,
                             desc='Variance of observation noise')


class BayesSerial(Bayes):
    def __init__(self, algorithm):
        n = len(algorithm.problem.parameters)
        super().__init__(n)

        # algorithm
        self.algorithm = algorithm

        # Size design variables.
        self.lb = np.empty((n,))
        self.ub = np.empty((n,))
        self.params = {}

        self.job = Job(self.algorithm.problem)

    def evaluateSample(self, x):
        return self.job.evaluate_scalar(x)

# class BayesOptSerial(BayesOpt):
#     """ BayesOpt algorithms """
#
#     def __init__(self, problem: Problem, name="BayesOpt"):
#         super().__init__(problem, name)
#
#         self.bo = BayesOptClassSerial(self)
#
#     def run(self):
#         # Figure out bounds vectors.
#         i = 0
#         for id in self.problem.parameters:
#             bounds = self.problem.parameters[id]['bounds']
#
#             self.bo.lb[i] = bounds[0]
#             self.bo.ub[i] = bounds[1]
#
#             i += 1
#
#         # set bayesopt
#         self.bo.params['n_iterations'] = self.options['n_iterations']
#         self.bo.params['n_init_samples'] = self.options['n_init_samples']
#         self.bo.params['n_iter_relearn'] = self.options['n_iter_relearn']
#         self.bo.params['surr_name'] = self.options['surr_name']
#         self.bo.params['surr_noise'] = self.options['surr_noise']
#         self.bo.params['init_method'] = self.options['init_method']
#         self.bo.params['l_type'] = self.options['l_type']
#         self.bo.params['sc_type'] = self.options['sc_type']
#         self.bo.params['verbose_level'] = self.options['verbose_level']
#
#         t_s = time.time()
#         self.problem.logger.info("BayesOpt: surr_name{}".format(self.options['surr_name']))
#         mvalue, x_out, error = self.bo.optimize()
#         t = time.time() - t_s
#         self.problem.logger.info("BayesOpt: elapsed time: {} s".format(t))
#
#         if error != 0:
#             print('Optimization FAILED.')
#             print("Error", error)
#             print('-' * 35)
#         else:
#             pass
#             # print('Optimization Complete, %f seconds' % (clock() - start))
#             # print("Result", x_out, mvalue)
#             # print('-' * 35)

class Queue:
    def __init__(self):
        self._queue = []

    @property
    def empty(self):
        return len(self) == 0

    def __len__(self):
        return len(self._queue)

    def __next__(self):
        if self.empty:
            raise StopIteration("Queue is empty, no more objects to retrieve.")
        obj = self._queue[0]
        self._queue = self._queue[1:]
        return obj

    def next(self):
        return self.__next__()

    def add(self, obj):
        """Add object to end of queue."""
        self._queue.append(obj)


# class Observable(object):
#     """
#
#     Inspired/Taken from
#         https://www.protechtraining.com/blog/post/879#simple-observer
#     """
#     def __init__(self, events):
#         # maps event names to subscribers
#         # str -> dict
#         self._events = {event: dict() for event in events}
#
#     def get_subscribers(self, event):
#         return self._events[event]
#
#     def subscribe(self, event, subscriber, callback=None):
#         if callback == None:
#             callback = getattr(subscriber, 'update')
#         self.get_subscribers(event)[subscriber] = callback
#
#     def unsubscribe(self, event, subscriber):
#         del self.get_subscribers(event)[subscriber]
#
#     def dispatch(self, event):
#         for _, callback in self.get_subscribers(event).items():
#             callback(event, self)


def acq_max(ac, gp, y_max, bounds, random_state, n_warmup=100000, n_iter=250):
    """
    A function to find the maximum of the acquisition function

    It uses a combination of random sampling (cheap) and the 'L-BFGS-B'
    optimization method. First by sampling `n_warmup` (1e5) points at random,
    and then running L-BFGS-B from `n_iter` (250) random starting points.

    Parameters
    ----------
    :param ac:
        The acquisition function object that return its point-wise value.

    :param gp:
        A gaussian process fitted to the relevant data.

    :param y_max:
        The current maximum known value of the target function.

    :param bounds:
        The variables bounds to limit the search of the acq max.

    :param random_state:
        instance of np.RandomState random number generator

    :param n_warmup:
        number of times to randomly sample the aquisition function

    :param n_iter:
        number of times to run scipy.minimize

    Returns
    -------
    :return: x_max, The arg max of the acquisition function.
    """

    # Warm up with random points
    x_tries = random_state.uniform(bounds[:, 0], bounds[:, 1],
                                   size=(n_warmup, bounds.shape[0]))
    ys = ac(x_tries, gp=gp, y_max=y_max)
    x_max = x_tries[ys.argmax()]
    max_acq = ys.max()

    # Explore the parameter space more throughly
    x_seeds = random_state.uniform(bounds[:, 0], bounds[:, 1],
                                   size=(n_iter, bounds.shape[0]))
    for x_try in x_seeds:
        # Find the minimum of minus the acquisition function
        res = minimize(lambda x: -ac(x.reshape(1, -1), gp=gp, y_max=y_max),
                       x_try.reshape(1, -1),
                       bounds=bounds,
                       method="L-BFGS-B")

        # See if success
        if not res.success:
            continue

        # Store it if better than previous minimum(maximum).
        if max_acq is None or -res.fun[0] >= max_acq:
            x_max = res.x
            max_acq = -res.fun[0]

    # Clip output to make sure it lies within the bounds. Due to floating
    # point technicalities this is not always the case.
    return np.clip(x_max, bounds[:, 0], bounds[:, 1])


class UtilityFunction(object):
    """
    An object to compute the acquisition functions.
    """

    def __init__(self, kind, kappa, xi):
        """
        If UCB is to be used, a constant kappa is needed.
        """
        self.kappa = kappa

        self.xi = xi

        if kind not in ['ucb', 'ei', 'poi']:
            err = "The utility function " \
                  "{} has not been implemented, " \
                  "please choose one of ucb, ei, or poi.".format(kind)
            raise NotImplementedError(err)
        else:
            self.kind = kind

    def utility(self, x, gp, y_max):
        if self.kind == 'ucb':
            return self._ucb(x, gp, self.kappa)
        if self.kind == 'ei':
            return self._ei(x, gp, y_max, self.xi)
        if self.kind == 'poi':
            return self._poi(x, gp, y_max, self.xi)

    @staticmethod
    def _ucb(x, gp, kappa):
        mean, std = gp.predict(x, return_std=True)

        return mean + kappa * std

    @staticmethod
    def _ei(x, gp, y_max, xi):
        mean, std = gp.predict(x, return_std=True)

        z = (mean - y_max - xi)/std
        return (mean - y_max - xi) * norm.cdf(z) + std * norm.pdf(z)

    @staticmethod
    def _poi(x, gp, y_max, xi):
        mean, std = gp.predict(x, return_std=True)

        z = (mean - y_max - xi)/std
        return norm.cdf(z)


def ensure_rng(random_state=None):
    """
    Creates a random number generator based on an optional seed.  This can be
    an integer or another random state for a seeded rng, or None for an
    unseeded rng.
    """
    if random_state is None:
        random_state = np.random.RandomState()
    elif isinstance(random_state, int):
        random_state = np.random.RandomState(random_state)
    else:
        assert isinstance(random_state, np.random.RandomState)
    return random_state


class TargetSpace(object):
    """
    Holds the param-space coordinates (X) and target values (Y)
    Allows for constant-time appends while ensuring no duplicates are added

    Example
    -------
    >>> def target_func(p1, p2):
    >>>     return p1 + p2
    >>> pbounds = {'p1': (0, 1), 'p2': (1, 100)}
    >>> space = TargetSpace(target_func, pbounds, random_state=0)
    >>> x = space.random_points(1)[0]
    >>> y = space.register_point(x)
    >>> assert self.max_point()['max_val'] == y
    """
    def __init__(self, target_func, pbounds, random_state=None):
        """
        Parameters
        ----------
        target_func : function
            Function to be maximized.

        pbounds : dict
            Dictionary with parameters names as keys and a tuple with minimum
            and maximum values.

        random_state : int, RandomState, or None
            optionally specify a seed for a random number generator
        """
        self.random_state = ensure_rng(random_state)

        # The function to be optimized
        self.target_func = target_func

        # Get the name of the parameters
        self._keys = sorted(pbounds)
        # Create an array with parameters bounds
        self._bounds = np.array(
            [item[1] for item in sorted(pbounds.items(), key=lambda x: x[0])],
            dtype=np.float
        )

        # preallocated memory for X and Y points
        self._params = np.empty(shape=(0, self.dim))
        self._target = np.empty(shape=(0))

        # keep track of unique points we have seen so far
        self._cache = {}

    def __contains__(self, x):
        return tuple(map(float, x)) in self._cache
        # return _hashable(x) in self._cache

    def __len__(self):
        assert len(self._params) == len(self._target)
        return len(self._target)

    @property
    def empty(self):
        return len(self) == 0

    @property
    def params(self):
        return self._params

    @property
    def target(self):
        return self._target

    @property
    def dim(self):
        return len(self._keys)

    @property
    def keys(self):
        return self._keys

    @property
    def bounds(self):
        return self._bounds

    def params_to_array(self, params):
        try:
            assert set(params) == set(self.keys)
        except AssertionError:
            raise ValueError(
                "Parameters' keys ({}) do ".format(sorted(params)) +
                "not match the expected set of keys ({}).".format(self.keys)
            )
        return np.asarray([params[key] for key in self.keys])

    def array_to_params(self, x):
        try:
            assert len(x) == len(self.keys)
        except AssertionError:
            raise ValueError(
                "Size of array ({}) is different than the ".format(len(x)) +
                "expected number of parameters ({}).".format(len(self.keys))
            )
        return dict(zip(self.keys, x))

    def _as_array(self, x):
        try:
            x = np.asarray(x, dtype=float)
        except TypeError:
            x = self.params_to_array(x)

        x = x.ravel()
        try:
            assert x.size == self.dim
        except AssertionError:
            raise ValueError(
                "Size of array ({}) is different than the ".format(len(x)) +
                "expected number of parameters ({}).".format(len(self.keys))
            )
        return x

    def register(self, params, target):
        """
        Append a point and its target value to the known data.

        Parameters
        ----------
        x : ndarray
            a single point, with len(x) == self.dim

        y : float
            target function value

        Raises
        ------
        KeyError:
            if the point is not unique

        Notes
        -----
        runs in ammortized constant time

        Example
        -------
        >>> pbounds = {'p1': (0, 1), 'p2': (1, 100)}
        >>> space = TargetSpace(lambda p1, p2: p1 + p2, pbounds)
        >>> len(space)
        0
        >>> x = np.array([0, 0])
        >>> y = 1
        >>> space.add_observation(x, y)
        >>> len(space)
        1
        """
        x = self._as_array(params)
        if x in self:
            raise KeyError('Data point {} is not unique'.format(x))

        # Insert data into unique dictionary
        self._cache[tuple(map(float, x.ravel()))] = target

        self._params = np.concatenate([self._params, x.reshape(1, -1)])
        self._target = np.concatenate([self._target, [target]])

    def probe(self, params):
        """
        Evaulates a single point x, to obtain the value y and then records them
        as observations.

        Notes
        -----
        If x has been previously seen returns a cached value of y.

        Parameters
        ----------
        x : ndarray
            a single point, with len(x) == self.dim

        Returns
        -------
        y : float
            target function value.
        """
        x = self._as_array(params)

        try:
            target = self._cache[tuple(map(float, x))]
        except KeyError:
            params = dict(zip(self._keys, x))
            target = self.target_func(**params)
            self.register(x, target)
        return target

    def random_sample(self):
        """
        Creates random points within the bounds of the space.

        Returns
        ----------
        data: ndarray
            [num x dim] array points with dimensions corresponding to `self._keys`

        Example
        -------
        >>> target_func = lambda p1, p2: p1 + p2
        >>> pbounds = {'p1': (0, 1), 'p2': (1, 100)}
        >>> space = TargetSpace(target_func, pbounds, random_state=0)
        >>> space.random_points(1)
        array([[ 55.33253689,   0.54488318]])
        """
        # TODO: support integer, category, and basic scipy.optimize constraints
        data = np.empty((1, self.dim))
        for col, (lower, upper) in enumerate(self._bounds):
            data.T[col] = self.random_state.uniform(lower, upper, size=1)
        return data.ravel()

    def max(self):
        """Get maximum target value found and corresponding parametes."""
        try:
            res = {
                'target': self.target.max(),
                'params': dict(
                    zip(self.keys, self.params[self.target.argmax()])
                )
            }
        except ValueError:
            res = {}
        return res

    def res(self):
        """Get all target values found and corresponding parametes."""
        params = [dict(zip(self.keys, p)) for p in self.params]

        return [
            {"target": target, "params": param}
            for target, param in zip(self.target, params)
        ]

    def set_bounds(self, new_bounds):
        """
        A method that allows changing the lower and upper searching bounds

        Parameters
        ----------
        new_bounds : dict
            A dictionary with the parameter name and its new bounds
        """
        for row, key in enumerate(self.keys):
            if key in new_bounds:
                self._bounds[row] = new_bounds[key]


# class BayesianOptimization(Observable):
class BayesianOptimization(object):
    def __init__(self, f, pbounds, random_state=None, verbose=2):
        """"""
        self._random_state = ensure_rng(random_state)

        # Data structure containing the function to be optimized, the bounds of
        # its domain, and a record of the evaluations we have done so far
        self._space = TargetSpace(f, pbounds, random_state)

        # queue
        self._queue = Queue()

        # Internal GP regressor
        self._gp = GaussianProcessRegressor(
            kernel=Matern(nu=2.5),
            alpha=1e-6,
            normalize_y=True,
            n_restarts_optimizer=25,
            random_state=self._random_state,
        )

        self._verbose = verbose
        # super(BayesianOptimization, self).__init__(events=DEFAULT_EVENTS)
        super(BayesianOptimization, self).__init__()

    @property
    def space(self):
        return self._space

    @property
    def max(self):
        return self._space.max()

    @property
    def res(self):
        return self._space.res()

    def register(self, params, target):
        """Expect observation with known target"""
        self._space.register(params, target)
        # self.dispatch(Events.OPTMIZATION_STEP)

    def probe(self, params, lazy=True):
        """Probe target of x"""
        if lazy:
            self._queue.add(params)
        else:
            self._space.probe(params)
            # self.dispatch(Events.OPTMIZATION_STEP)

    def suggest(self, utility_function):
        """Most promissing point to probe next"""
        if len(self._space) == 0:
            return self._space.array_to_params(self._space.random_sample())

        # Sklearn's GP throws a large number of warnings at times, but
        # we don't really need to see them here.
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if len(self._space.params) % 5 == 0:
                self._gp.fit(self._space.params, self._space.target)

        # Finding argmax of the acquisition function.
        suggestion = acq_max(
            ac=utility_function.utility,
            gp=self._gp,
            y_max=self._space.target.max(),
            bounds=self._space.bounds,
            random_state=self._random_state
        )

        return self._space.array_to_params(suggestion)

    def _prime_queue(self, init_points):
        """Make sure there's something in the queue at the very beginning."""
        if self._queue.empty and self._space.empty:
            init_points = max(init_points, 1)

        for _ in range(init_points):
            self._queue.add(self._space.random_sample())

    def _prime_subscriptions(self):
        #if not any([len(subs) for subs in self._events.values()]):
        #    _logger = _get_default_logger(self._verbose)
        #    self.subscribe(Events.OPTMIZATION_START, _logger)
        #    self.subscribe(Events.OPTMIZATION_STEP, _logger)
        #    self.subscribe(Events.OPTMIZATION_END, _logger)
        pass

    def maximize(self,
                 init_points=5,
                 n_iter=25,
                 acq='ucb',
                 kappa=2.576,
                 xi=0.0,
                 **gp_params):
        """Mazimize your function"""
        self._prime_subscriptions()
        # self.dispatch(Events.OPTMIZATION_START)
        self._prime_queue(init_points)
        self.set_gp_params(**gp_params)

        util = UtilityFunction(kind=acq, kappa=kappa, xi=xi)
        iteration = 0
        while not self._queue.empty or iteration < n_iter:
            try:
                x_probe = next(self._queue)
            except StopIteration:
                x_probe = self.suggest(util)
                iteration += 1

            self.probe(x_probe, lazy=False)

        # self.dispatch(Events.OPTMIZATION_END)

    def set_bounds(self, new_bounds):
        """
        A method that allows changing the lower and upper searching bounds

        Parameters
        ----------
        new_bounds : dict
            A dictionary with the parameter name and its new bounds
        """
        self._space.set_bounds(new_bounds)

    def set_gp_params(self, **params):
        self._gp.set_params(**params)


from artap.problem import Problem
from artap.benchmark_functions import BinhAndKorn, AckleyN2
from artap.algorithm_genetic import NSGAII
# from artap.results import Results
from artap.results import GraphicalResults


class MyProblem(Problem):
    """ Describe simple one objective optimization problem. """
    def __init__(self, name):
        parameters = {'x_1': {'initial_value': 2.5, 'bounds': [0, 5]},
                      'x_2': {'initial_value': 1.5, 'bounds': [0, 3]}}
        costs = ['F_1', 'F_2']

        super().__init__(name, parameters, costs)

    def evaluate(self, x):
        function = BinhAndKorn()
        return function.eval(x)

    def evaluate_constraints(self, x):
        return BinhAndKorn.constraints(x)


problem = MyProblem("NSGA2Optimization")
algorithm = NSGAII(problem)
algorithm.options['max_population_number'] = 30
algorithm.options['max_population_size'] = 100
# algorithm.options['calculate_gradients'] = True
algorithm.run()

# results = GraphicalResults(problem)
# results.plot_scatter('F_1', 'F_2', filename="/tmp/scatter.pdf")
# results.plot_scatter('x_1', 'x_2')
# results.plot_individuals('F_1')



import platform
from abc import *
from random import uniform
from enum import Enum


class Individual(metaclass=ABCMeta):
    """ Collects information about one point in design space. """

    class State(Enum):
        EMPTY = 0
        IN_PROGRESS = 1
        EVALUATED = 2
        FAILED = 3

        @staticmethod
        def to_string(state):
            if state == Individual.State.EMPTY:
                return "Individual is empty."
            elif state == Individual.State.IN_PROGRESS:
                return "Individual is in progress."
            elif state == Individual.State.EVALUATED:
                return "Individual is evaluated."
            elif state == Individual.State.FAILED:
                return "Individual evaluation failed."

    def __init__(self, vector: list):
        self.vector = vector.copy()
        self.costs = []
        self.custom = {}
        self.state = self.State.EMPTY
        self.info = {"start_time": 0.0,
                     "finish_time": 0.0,
                     "population": -1,
                     "processor": platform.processor(),
                     "system": platform.system(),
                     "python": platform.python_version(),
                     "hostname": platform.node()}

    def __repr__(self):
        """ :return: [vector[p1, p2, ... pn]; costs[c1, c2, ... cn]] """

        string = "vector: ["
        for i, number in enumerate(self.vector):
            string += str(number)
            if i < len(self.vector) - 1:
                string += ", "
        string = string[:len(string) - 1]
        string += "]"

        string += "; costs:["
        for i, number in enumerate(self.costs):
            string += str(number)
            if i < len(self.costs) - 1:
                string += ", "
        string += "]"

        if len(self.custom) > 0:
            string += "; custom:["
            for i, number in enumerate(self.custom):
                string += str(number)
                if i < len(self.custom) - 1:
                    string += ", "
            string += "]"

        string += "; state: '{}', ".format(Individual.State.to_string(self.state))
        string += "; info: {}, ".format(self.info)

        return string

    def __eq__(self, other):
        return self.vector == other.vector and self.costs == self.costs and self.custom == self.custom and self.state == self.state

    def __hash__(self):
        return id(self)

    def sync(self, individual):
        self.vector = individual.vector
        self.costs = individual.costs
        self.custom = individual.custom
        self.state = individual.state
        self.info = individual.info


class GeneticIndividual(Individual):
    def __init__(self, vector: list):
        super().__init__(vector)

        self.gradient = []
        self.feasible = 0.0  # the distance from the feasibility region in min norm, its an index, not a
        self.dominate = set()
        self.domination_counter = 0
        self.front_number = None  # TODO: make better
        self.crowding_distance = 0  # TODO: deprecated? --
        self.depends_on = None  # For calculating gradients
        self.modified_param = -1
        # For particle swarm optimization
        self.velocity_i = [0] * len(vector)  # particle velocity
        self.best_parameters = []  # best position individual
        self.best_costs = []  # best error individual

        # transformed data, like a cache for the signed values of the calculation results in ndarray
        self.signed_costs = []

        for i in range(0, len(self.vector)):
            self.velocity_i.append(uniform(-1, 1))

    def transform_data(self, signs):
        """
        Tha aim of this function is to transform the data into a numpy array, which can make a faster
        comparison during the sorting of the data.

        :param signs: this is the list of the signs of every value, which shows that the minimum or the maximum
        of the values are calculated. This list is stored in the Problem class

        :return: nothing but calculates np array from the data
        """

        # svec = asarray(signs,dtype=float64)
        # cvec = asarray(self.costs, dtype=float64)

        # self.signed_costs = svec*cvec
        # self.signed_costs = append(self.signed_costs, self.feasible)

        self.signed_costs = list(map(lambda x, y: x * y, signs, self.costs))
        self.signed_costs.append(self.feasible)
        return

    def weighted_cost(self, weights):
        """
        Calculates a single best solution

        :param weights:
        :return:
        """

        return

    def __repr__(self):
        """ :return: [vector[p1, p2, ... pn]; costs[c1, c2, ... cn]] """

        string = "{}, ".format(super().__repr__())

        string += ", front number: {}".format(self.front_number)
        string += ", crowding distance: {}".format(self.crowding_distance)
        string += ", gradient: ["
        for i, number in enumerate(self.gradient):
            string += str(number)
            if i < len(self.vector) - 1:
                string += ", "
        string = string[:len(string) - 1]
        string += "]"

        return string

    def sync(self, individual):
        super().sync(individual)

        self.gradient = individual.gradient
        self.feasible = individual.feasible
        self.dominate = individual.dominate
        self.domination_counter = individual.domination_counter
        self.front_number = individual.front_number
        self.crowding_distance = individual.crowding_distance
        self.depends_on = individual.depends_on
        self.modified_param = individual.modified_param
        # For particle swarm optimization
        self.velocity_i = individual.velocity_i
        self.best_parameters = individual.best_parameters
        self.best_costs = individual.best_costs
